import unittest
import os.path

from tableaudocumentapi import Datasource
from lxml import etree as ET


TEST_ASSET_DIR = os.path.join(
    os.path.dirname(__file__),
    'assets'
)
TEST_TDS_FILE = os.path.join(
    TEST_ASSET_DIR,
    'field_change_test.tds'
)
TEST_TDS_FILE_OUTPUT = os.path.join(
    TEST_ASSET_DIR,
    'field_change_test_output.tds'
)

MESSAGES = {
    'test_change_values1': 'Value has not changed when altering values for {}.',
    'test_change_values2': 'XML-Structure has not changed when altering values for {}.',
    'test_change_valuesFail1': 'Value has changed when submitting the wrong value for {}.',
    'test_change_valuesFail2': 'XML-Structure has changed when submitting the wrong value for {}.',
    'test_change_aliases1': 'XML-Structure has not changed when altering aliases for {}.',
    'test_change_aliases2': 'Values have not changed when altering aliases for {}.'

}

NEW_VALUES = {
    'caption': 'testcaption',
    'alias': 'testalias',
    'datatype': 'boolean',
    'role': 'measure',
    'type': 'ordinal'
}

WRONG_VALUES = {
    'datatype': 'boolani',
    'role': 'messhure',
    'type': 'gordinol'
}

ALIASES = {
    'one': 'two',
    'three': 'four',
    'five': 'six'
}


class TestFieldChange(unittest.TestCase):

    def setUp(self):
        self.tds = Datasource.from_file(TEST_TDS_FILE)

    def current_hash(self):
        """ Return a hash of the current state of the XML.

        Allows us to easily identify whether the underlying XML-structure
        of a TDS-file has actually changed. Avoids false positives if,
        for example, a fields value has changed but the XML hasn't.
        """
        return hash(ET.tostring(self.tds._datasourceTree.getroot()))

    def check_state_change(self, should_change, msg, field_name):
        """ Check whether the XML has changed and update the current state.

            Args:
                should_change: Whether the XML is supposed to have changed or not. Boolean.
                msg: The message to be displayed in an error case, as key for the MESSAGES dict. String.
                field_name: The field name that will be displayed in the error message. String.

            Returns:
                Nothing.
        """
        new_state = self.current_hash()
        compare_func = self.assertNotEqual if should_change else self.assertEqual
        compare_func(
            self.state,
            new_state,
            msg=MESSAGES[msg].format(field_name)
        )
        self.state = new_state

    def test_change_values(self):
        """ Test if the value changes of a field are reflected in the object and in the underlying XML structure.
        """
        field_to_test = "[amount]"
        self.state = self.current_hash()
        # change all fields
        for key, value in NEW_VALUES.items():
            setattr(self.tds.fields[field_to_test], key, value)
            # the new value must be reflected in the object
            self.assertEqual(
                getattr(self.tds.fields[field_to_test], key),
                value,
                msg=MESSAGES['test_change_values1'].format(key)
            )
            # the new value must be reflected in the xml
            self.check_state_change(True, 'test_change_values2', key)

    def test_change_values_fail(self):
        """ Test if the value changes of a field are rejected if the wrong arguments are passed.
        """
        field_to_test = "[amount]"
        self.state = self.current_hash()
        # change all fields
        for key, value in WRONG_VALUES.items():

            with self.assertRaises(ValueError):
                # this must fail
                setattr(self.tds.fields[field_to_test], key, value)

            # the new value must NOT be reflected in the object
            self.assertNotEqual(
                getattr(self.tds.fields[field_to_test], key),
                value,
                msg=MESSAGES['test_change_valuesFail1'].format(key)
            )
            # the new value must NOT be reflected in the xml
            self.check_state_change(False, 'test_change_valuesFail2', key)

    def test_remove_field(self):
        """ Test if a Field can be removed.
        """
        field_to_test = "[amount]"
        self.state = self.current_hash()
        # change all fields
        field = self.tds.fields["[amount]"]
        self.tds.remove_field(field)
        self.assertNotEqual(self.state, self.current_hash())

    def test_change_aliases(self):
        """ Test if the alias changes of a field are reflected in the object and in the underlying XML structure.
        """
        field_to_test = "[amount]"
        self.state = self.current_hash()
        # change all fields
        for key, value in ALIASES.items():
            self.tds.fields[field_to_test].add_alias(key, value)
            # the new value must be reflected in the xml
            self.check_state_change(True, 'test_change_aliases1', field_to_test)

        # check whether all fields of ALIASES have been applied
        self.assertEqual(
            set(self.tds.fields[field_to_test].aliases),
            set(ALIASES),
            msg=MESSAGES['test_change_aliases2'].format(field_to_test)
        )

    def test_calculation_base(self):
        """ Test if the initial state of calculated fields is correct.
        """
        # Demo data has 2 calculated fields at the start
        original_len = len(self.tds.calculations)

        # Can't write to calculation for not-calculated fields!
        self.tds.fields['[name]'].calculation = '1 * 2'
        self.assertEqual(len(self.tds.calculations), original_len + 1)
        self.tds.fields['[name]'].calculation = '2 * 3'
        self.assertEqual(len(self.tds.calculations), original_len + 1)
        self.tds.fields['[price]'].calculation = '2 * 3'
        self.assertEqual(len(self.tds.calculations), original_len + 2)

    def test_calculation_change(self):
        """ Test whether changing calculations of a field works.
        """
        self.state = self.current_hash()
        new_calc = '33 * 44'
        fld_name = '[Calculation_357754699576291328]'
        self.tds.calculations[fld_name].calculation = new_calc

        # Check object representation
        self.assertEqual(self.tds.calculations[fld_name].calculation, new_calc)

        # Check XML representation
        new_state = self.current_hash()
        self.assertNotEqual(self.state, new_state)

    def test_calculation_new(self):
        """ Test if creating a new calculation works.
        """
        args = 'TestCalc', '12*34', 'integer', 'measure', 'quantitative', 'False'
        original_len = len(self.tds.calculations)
        self.tds.add_calculation(*args)
        self.assertEqual(len(self.tds.calculations), original_len + 1)

    def test_calculation_remove(self):
        """ Test if deleting a calculation works.
        """
        args = 'TestCalc2', '12*34', 'integer', 'measure', 'quantitative', 'True'

        original_len = len(self.tds.calculations)
        calc = self.tds.add_calculation(*args)
        self.assertEqual(len(self.tds.calculations), original_len + 1)

        self.tds.remove_field(calc)
        self.assertEqual(len(self.tds.calculations), original_len)

    def tearDown(self):
        """ Test if the file can be saved.
        Output file will be ignored by git, but can be used to verify the results.
        """
        self.tds.save_as(TEST_TDS_FILE_OUTPUT)


if __name__ == '__main__':
    unittest.main()



import unittest
import os.path

from tableaudocumentapi import Datasource, Field
from tableaudocumentapi.field import _find_metadata_record

TEST_ASSET_DIR = os.path.join(
    os.path.dirname(__file__),
    'assets'
)
TEST_TDS_FILE = os.path.join(
    TEST_ASSET_DIR,
    'datasource_test.tds'
)
TEST_UNICODE_FILE = os.path.join(
    TEST_ASSET_DIR,
    'unicode.tds'
)


class FieldsUnitTest(unittest.TestCase):
    def test_field_throws_if_no_data_passed_in(self):
        with self.assertRaises(AttributeError):
            Field()


class FindMetaDataRecordEdgeTest(unittest.TestCase):
    class MockXmlWithNoFind(object):
        def find(self, *args, **kwargs):
            return None

    def test_find_metadata_record_returns_none(self):
        self.assertIsNone(_find_metadata_record(self.MockXmlWithNoFind(), 'foo'))


class FieldsHandleUnicode(unittest.TestCase):
    def test_description_unicode(self):
        ds = Datasource.from_file(TEST_UNICODE_FILE)
        self.assertIsNotNone(ds.fields['A'].description)



from lxml import etree as ET
from tableaudocumentapi.dbclass import is_valid_dbclass


class Connection(object):
    """A class representing connections inside Data Sources."""

    def __init__(self, connxml):
        """Connection is usually instantiated by passing in connection elements
        in a Data Source. If creating a connection from scratch you can call
        `from_attributes` passing in the connection attributes.

        """
        self._connectionXML = connxml
        self._dbname = connxml.get('dbname')
        self._server = connxml.get('server')
        self._username = connxml.get('username')
        self._authentication = connxml.get('authentication')
        self._class = connxml.get('class')
        self._schema = connxml.get('schema')
        self._service = connxml.get('service')
        self._port = connxml.get('port', None)
        self._query_band = connxml.get('query-band-spec', None)
        self._initial_sql = connxml.get('one-time-sql', None)

    def __repr__(self):
        return "'<Connection server='{}' dbname='{}' @ {}>'".format(self._server, self._dbname, hex(id(self)))

    @classmethod
    def from_attributes(cls, server, dbname, username, dbclass, port=None, query_band=None,
                        initial_sql=None, authentication='', schema='', service=''):
        """Creates a new connection that can be added into a Data Source.
        defaults to `''` which will be treated as 'prompt' by Tableau."""

        root = ET.Element('connection', authentication=authentication)
        xml = cls(root)
        xml.server = server
        xml.dbname = dbname
        xml.username = username
        xml.schema = schema
        xml.service = service
        xml.dbclass = dbclass
        xml.port = port
        xml.query_band = query_band
        xml.initial_sql = initial_sql

        return xml

    @property
    def dbname(self):
        """Database name for the connection. Not the table name."""
        return self._dbname

    @dbname.setter
    def dbname(self, value):
        """
        Set the connection's database name property.

        Args:
            value:  New name of the database. String.

        Returns:
            Nothing.

        """
        self._dbname = value
        self._connectionXML.set('dbname', value)

    @property
    def server(self):
        """Hostname or IP address of the database server. May also be a URL in some connection types."""
        return self._server

    @server.setter
    def server(self, value):
        """
        Set the connection's server property.

        Args:
            value:  New server. String.

        Returns:
            Nothing.

        """
        self._server = value
        self._connectionXML.set('server', value)

    @property
    def username(self):
        """Username used to authenticate to the database."""
        return self._username

    @username.setter
    def username(self, value):
        """
        Set the connection's username property.

        Args:
            value:  New username value. String.

        Returns:
            Nothing.

        """
        self._username = value
        self._connectionXML.set('username', value)

    @property
    def authentication(self):
        return self._authentication

    @property
    def dbclass(self):
        """The type of connection (e.g. 'MySQL', 'Postgresql'). A complete list
        can be found in dbclass.py"""
        return self._class

    @dbclass.setter
    def dbclass(self, value):
        """Set the connection's dbclass property.

        Args:
            value:  New dbclass value. String.

        Returns:
            Nothing.
        """

        if not is_valid_dbclass(value):
            raise AttributeError("'{}' is not a valid database type".format(value))

        self._class = value
        self._connectionXML.set('class', value)

    @property
    def port(self):
        """Port used to connect to the database."""
        return self._port

    @port.setter
    def port(self, value):
        """Set the connection's port property.

        Args:
            value:  New port value. String.

        Returns:
            Nothing.
        """

        self._port = value
        # If port is None we remove the element and don't write it to XML
        if value is None:
            try:
                del self._connectionXML.attrib['port']
            except KeyError:
                pass
        else:
            self._connectionXML.set('port', value)

    @property
    def query_band(self):
        """Query band passed on connection to database."""
        return self._query_band

    @query_band.setter
    def query_band(self, value):
        """Set the connection's query_band property.

        Args:
            value:  New query_band value. String.

        Returns:
            Nothing.
        """

        self._query_band = value
        # If query band is None we remove the element and don't write it to XML
        if value is None:
            try:
                del self._connectionXML.attrib['query-band-spec']
            except KeyError:
                pass
        else:
            self._connectionXML.set('query-band-spec', value)

    @property
    def initial_sql(self):
        """Initial SQL to be run."""
        return self._initial_sql

    @initial_sql.setter
    def initial_sql(self, value):
        """Set the connection's initial_sql property.

        Args:
            value:  New initial_sql value. String.

        Returns:
            Nothing.
        """

        self._initial_sql = value
        # If initial_sql is None we remove the element and don't write it to XML
        if value is None:
            try:
                del self._connectionXML.attrib['one-time-sql']
            except KeyError:
                pass
        else:
            self._connectionXML.set('one-time-sql', value)

    @property
    def schema(self):
        """Database schema for the connection. Not the table name."""
        return self._schema

    @schema.setter
    def schema(self, value):
        """
        Set the connection's schema property.

        Args:
            value:  New name of the database schema. String.

        Returns:
            Nothing.

        """
        self._schema = value
        self._connectionXML.set('schema', value)

    @property
    def service(self):
        """Database service for the connection. Not the table name."""
        return self._service

    @service.setter
    def service(self, value):
        """
        Set the connection's service property.

        Args:
            value:  New name of the database service. String.

        Returns:
            Nothing.

        """
        self._service = value
        self._connectionXML.set('service', value)



import contextlib
import os
import shutil
import tempfile
import zipfile
from lxml import etree as ET

from distutils.version import LooseVersion as Version

MIN_SUPPORTED_VERSION = Version("9.0")


class TableauVersionNotSupportedException(Exception):
    pass


class TableauInvalidFileException(Exception):
    pass


def xml_open(filename, expected_root=None):
    """Opens the provided 'filename'. Handles detecting if the file is an archive,
    detecting the document version, and validating the root tag."""

    # Is the file a zip (.twbx or .tdsx)
    if zipfile.is_zipfile(filename):
        tree = get_xml_from_archive(filename)
    else:
        _register_all_namespaces()
        tree = ET.parse(filename)

    # Is the file a supported version
    tree_root = tree.getroot()
    file_version = Version(tree_root.attrib.get('version', '0.0'))

    if file_version < MIN_SUPPORTED_VERSION:
        raise TableauVersionNotSupportedException(file_version)

    # Does the root tag match the object type (workbook or data source)
    if expected_root and (expected_root != tree_root.tag):
        if expected_root == 'workbook' and tree_root.tag == 'datasource':
            return  # A .twbx can contain .tds files if it contains custom geocoding.
        raise TableauInvalidFileException(
            "'{}'' is not a valid '{}' file".format(filename, expected_root))

    return tree


@contextlib.contextmanager
def temporary_directory(*args, **kwargs):
    d = tempfile.mkdtemp(*args, **kwargs)
    try:
        yield d
    finally:
        shutil.rmtree(d)


def _register_all_namespaces():
    # TO DO: should look at the file to find namespaces, not hardcode this one
    ET.register_namespace("user", "http://www.tableausoftware.com/xml/user")


def find_file_in_zip(zip_file):
    '''Returns the twb/tds file from a Tableau packaged file format. Packaged
    files can contain cache entries which are also valid XML, so only look for
    files with a .tds or .twb extension.
    '''

    candidate_files = filter(lambda x: x.split('.')[-1] in ('twb', 'tds'),
                             zip_file.namelist())

    for filename in candidate_files:
        with zip_file.open(filename) as xml_candidate:
            try:
                ET.parse(xml_candidate)
                return filename
            except ET.ParseError:
                # That's not an XML file by gosh
                pass


def get_xml_from_archive(filename):
    with zipfile.ZipFile(filename, allowZip64=True) as zf:
        with zf.open(find_file_in_zip(zf)) as xml_file:
            xml_tree = ET.parse(xml_file)

    return xml_tree


def build_archive_file(archive_contents, zip_file):
    """Build a Tableau-compatible archive file."""

    # This is tested against Desktop and Server, and reverse engineered by lots
    # of trial and error. Do not change this logic.
    for root_dir, _, files in os.walk(archive_contents):
        relative_dir = os.path.relpath(root_dir, archive_contents)
        for f in files:
            temp_file_full_path = os.path.join(
                archive_contents, relative_dir, f)
            zipname = os.path.join(relative_dir, f)
            zip_file.write(temp_file_full_path, arcname=zipname)


def save_into_archive(xml_tree, filename, new_filename=None):
    # Saving an archive means extracting the contents into a temp folder,
    # saving the changes over the twb/tds in that folder, and then
    # packaging it back up into a zip with a very specific format
    # e.g. no empty files for directories, which Windows and Mac do by default

    if new_filename is None:
        new_filename = filename

    # Extract to temp directory
    with temporary_directory() as temp_path:
        with zipfile.ZipFile(filename, allowZip64=True) as zf:
            xml_file = find_file_in_zip(zf)
            zf.extractall(temp_path)
        # Write the new version of the file to the temp directory
        xml_tree.write(os.path.join(
            temp_path, xml_file), encoding="utf-8", xml_declaration=True)

        # Write the new archive with the contents of the temp folder
        with zipfile.ZipFile(new_filename, "w", compression=zipfile.ZIP_DEFLATED, allowZip64=True) as new_archive:
            build_archive_file(temp_path, new_archive)


def _save_file(container_file, xml_tree, new_filename=None):

    _register_all_namespaces()  # this shouldn't be necessary, should be done on open

    if new_filename is None:
        new_filename = container_file

    if zipfile.is_zipfile(container_file):
        save_into_archive(xml_tree, container_file, new_filename)
    else:
        xml_tree.write(new_filename, encoding="utf-8", xml_declaration=True)



import unittest
import os.path
import functools

from tableaudocumentapi.multilookup_dict import MultiLookupDict


class MLDTests(unittest.TestCase):
    def setUp(self):
        self.mld = MultiLookupDict({
            '[foo]': {
                'alias': 'bar',
                'caption': 'baz',
                'value': 1
            },
            '[bar]': {
                'caption': 'foo',
                'value': 2
            },
            '[baz]': {
                'value': 3
            }
        })

    def test_multilookupdict_can_be_empty(self):
        mld = MultiLookupDict()
        self.assertIsNotNone(mld)

    def test_multilookupdict_name_only(self):
        actual = self.mld['[baz]']
        self.assertEqual(3, actual['value'])

    def test_multilookupdict_alias_overrides_everything(self):
        actual = self.mld['bar']
        self.assertEqual(1, actual['value'])

    def test_mutlilookupdict_caption_overrides_id(self):
        actual = self.mld['foo']
        self.assertEqual(2, actual['value'])

    def test_mutlilookupdict_can_still_find_id_even_with_alias(self):
        actual = self.mld['[foo]']
        self.assertEqual(1, actual['value'])

    def test_mutlilookupdict_can_still_find_caption_even_with_alias(self):
        actual = self.mld['baz']
        self.assertEqual(1, actual['value'])

    def test_mutlilookupdict_can_still_find_id_even_with_caption(self):
        actual = self.mld['[bar]']
        self.assertEqual(2, actual['value'])

    def test_multilookupdict_gives_key_error_on_invalid_key(self):
        try:
            self.mld.get('foobar')
            self.fail('should have thrown key error')
        except KeyError as ex:
            self.assertEqual(str(ex), "'foobar'")

    def test_multilookupdict_get_returns_default_value(self):
        default_value = ('default', 'return', 'value')
        actual = self.mld.get('foobar', default_value)
        self.assertEqual(actual, default_value)

    def test_multilookupdict_get_returns_value(self):
        actual = self.mld.get('baz')
        self.assertEqual(1, actual['value'])

    def test_multilookupdict_can_set_item(self):
        before = self.mld['baz']
        self.mld['baz'] = 4
        self.assertEqual(4, self.mld['baz'])

    def test_multilookupdict_can_set_new_item(self):
        self.mld['wakka'] = 1
        self.assertEqual(1, self.mld['wakka'])

    def test_multilookupdict_can_set_with_alias(self):
        self.mld['bar'] = 2
        self.assertEqual(2, self.mld['[foo]'])



import weakref

from tableaudocumentapi import Datasource, xfile
from tableaudocumentapi.xfile import xml_open, TableauInvalidFileException


class Workbook(object):
    """A class for writing Tableau workbook files."""

    def __init__(self, filename):
        """Open the workbook at `filename`. This will handle packaged and unpacked
        workbook files automatically. This will also parse Data Sources and Worksheets
        for access.

        """

        self._filename = filename

        self._workbookTree = xml_open(self._filename, 'workbook')
        if not self._workbookTree:
            raise TableauInvalidFileException("Workbook file must have a workbook element at root")

        self._workbookRoot = self._workbookTree.getroot()

        self._dashboards = self._prepare_dashboards(self._workbookRoot)

        self._datasources = self._prepare_datasources(
            self._workbookRoot)

        self._datasource_index = self._prepare_datasource_index(self._datasources)

        self._worksheets = self._prepare_worksheets(
            self._workbookRoot, self._datasource_index)

        self._shapes = self._prepare_shapes(self._workbookRoot)

    @property
    def dashboards(self):
        return self._dashboards

    @property
    def datasources(self):
        return self._datasources

    @property
    def worksheets(self):
        return self._worksheets

    @property
    def filename(self):
        return self._filename

    @property
    def shapes(self):
        return self._shapes

    def save(self):
        """
        Call finalization code and save file.

        Args:
            None.

        Returns:
            Nothing.

        """

        # save the file
        xfile._save_file(self._filename, self._workbookTree)

    def save_as(self, new_filename):
        """
        Save our file with the name provided.

        Args:
            new_filename:  New name for the workbook file. String.

        Returns:
            Nothing.

        """
        xfile._save_file(
            self._filename, self._workbookTree, new_filename)

    @staticmethod
    def _prepare_datasource_index(datasources):
        retval = weakref.WeakValueDictionary()
        for datasource in datasources:
            retval[datasource.name] = datasource

        return retval

    @staticmethod
    def _prepare_datasources(xml_root):
        datasources = []

        # loop through our datasources and append
        datasource_elements = xml_root.find('datasources')
        if datasource_elements is None:
            return []

        for datasource in datasource_elements:
            ds = Datasource(datasource)
            datasources.append(ds)

        return datasources

    @staticmethod
    def _prepare_dashboards(xml_root):
        dashboards = []

        dashboard_elements = xml_root.find('.//dashboards')
        if dashboard_elements is None:
            return []

        for dash_element in dashboard_elements:
            dash_name = dash_element.attrib['name']
            dashboards.append(dash_name)

        return dashboards

    @staticmethod
    def _prepare_worksheets(xml_root, ds_index):
        worksheets = []
        worksheets_element = xml_root.find('.//worksheets')
        if worksheets_element is None:
            return worksheets

        for worksheet_element in worksheets_element:
            worksheet_name = worksheet_element.attrib['name']
            worksheets.append(worksheet_name)  # TODO: A real worksheet object, for now, only name

            dependencies = worksheet_element.findall('.//datasource-dependencies')

            for dependency in dependencies:
                datasource_name = dependency.attrib['datasource']
                datasource = ds_index[datasource_name]
                for column in dependency.findall('.//column'):
                    column_name = column.attrib['name']
                    if column_name in datasource.fields:
                        datasource.fields[column_name].add_used_in(worksheet_name)

        return worksheets

    @staticmethod
    def _prepare_shapes(xml_root):
        shapes = []
        worksheets_element = xml_root.find('.//external/shapes')
        if worksheets_element is None:
            return shapes

        for worksheet_element in worksheets_element:
            shape_name = worksheet_element.attrib['name']
            shapes.append(shape_name)

        return shapes



import collections
import itertools
from lxml import etree as ET
import xml.sax.saxutils as sax
from uuid import uuid4

from tableaudocumentapi import Connection, xfile
from tableaudocumentapi import Field
from tableaudocumentapi.multilookup_dict import MultiLookupDict
from tableaudocumentapi.xfile import xml_open


_ColumnObjectReturnTuple = collections.namedtuple('_ColumnObjectReturnTupleType', ['id', 'object'])


def _get_metadata_xml_for_field(root_xml, field_name):
    if "'" in field_name:
        field_name = sax.escape(field_name, {"'": "&apos;"})
    xpath = u".//metadata-record[@class='column'][local-name='{}']".format(field_name)
    return root_xml.find(xpath)


def _is_used_by_worksheet(names, field):
    return any(y for y in names if y in field.worksheets)


class FieldDictionary(MultiLookupDict):

    def used_by_sheet(self, name):
        # If we pass in a string, no need to get complicated, just check to see if name is in
        # the field's list of worksheets
        if isinstance(name, str):
            return [x for x in self.values() if name in x.worksheets]

        # if we pass in a list, we need to check to see if any of the names in the list are in
        # the field's list of worksheets
        return [x for x in self.values() if _is_used_by_worksheet(name, x)]


def _column_object_from_column_xml(root_xml, column_xml):
    field_object = Field.from_column_xml(column_xml)
    local_name = field_object.id
    metadata_record = _get_metadata_xml_for_field(root_xml, local_name)
    if metadata_record is not None:
        field_object.apply_metadata(metadata_record)
    return _ColumnObjectReturnTuple(field_object.id, field_object)


def _column_object_from_metadata_xml(metadata_xml):
    field_object = Field.from_metadata_xml(metadata_xml)
    return _ColumnObjectReturnTuple(field_object.id, field_object)


def base36encode(number):
    """Converts an integer into a base36 string."""

    ALPHABET = "0123456789abcdefghijklmnopqrstuvwxyz"

    base36 = ''
    sign = ''

    if number < 0:
        sign = '-'
        number = -number

    if 0 <= number < len(ALPHABET):
        return sign + ALPHABET[number]

    while number != 0:
        number, i = divmod(number, len(ALPHABET))
        base36 = ALPHABET[i] + base36

    return sign + base36


def _make_unique_name(dbclass):
    rand_part = base36encode(uuid4().int)
    name = dbclass + '.' + rand_part
    return name


class ConnectionParser(object):
    """Parser for detecting and extracting connections from differing Tableau file formats."""

    def __init__(self, datasource_xml, version):
        self._dsxml = datasource_xml
        self._dsversion = version

    def _extract_federated_connections(self):
        connections = list(map(Connection, self._dsxml.findall('.//named-connections/named-connection/*')))
        # 'sqlproxy' connections (Tableau Server Connections) are not embedded into named-connection elements
        # extract them manually for now
        connections.extend(map(Connection, self._dsxml.findall("./connection[@class='sqlproxy']")))
        return connections

    def _extract_legacy_connection(self):
        return list(map(Connection, self._dsxml.findall('connection')))

    def get_connections(self):
        """Find and return all connections based on file format version."""

        if float(self._dsversion) < 10:
            connections = self._extract_legacy_connection()
        else:
            connections = self._extract_federated_connections()
        return connections


class Datasource(object):
    """A class representing Tableau Data Sources, embedded in workbook files or
    in TDS files.

    """

    def __init__(self, dsxml, filename=None):
        """
        Constructor.  Default is to create datasource from xml.

        """
        self._filename = filename
        self._datasourceXML = dsxml
        self._datasourceTree = ET.ElementTree(self._datasourceXML)
        self._name = self._datasourceXML.get('name') or self._datasourceXML.get(
            'formatted-name')  # TDS files don't have a name attribute
        self._version = self._datasourceXML.get('version')
        self._caption = self._datasourceXML.get('caption', '')
        self._connection_parser = ConnectionParser(
            self._datasourceXML, version=self._version)
        self._connections = self._connection_parser.get_connections()
        self._fields = None

    @classmethod
    def from_file(cls, filename):
        """Initialize datasource from file (.tds ot .tdsx)"""

        dsxml = xml_open(filename, 'datasource').getroot()
        return cls(dsxml, filename)

    @classmethod
    def from_connections(cls, caption, connections):
        """Create a new Data Source give a list of Connections."""

        root = ET.Element('datasource', caption=caption, version='10.0', inline='true')
        outer_connection = ET.SubElement(root, 'connection')
        outer_connection.set('class', 'federated')
        named_conns = ET.SubElement(outer_connection, 'named-connections')
        for conn in connections:
            nc = ET.SubElement(named_conns,
                               'named-connection',
                               name=_make_unique_name(conn.dbclass),
                               caption=conn.server)
            nc.append(conn._connectionXML)
        return cls(root)

    def save(self):
        """
        Call finalization code and save file.

        Args:
            None.

        Returns:
            Nothing.

        """

        # save the file

        xfile._save_file(self._filename, self._datasourceTree)

    def save_as(self, new_filename):
        """
        Save our file with the name provided.

        Args:
            new_filename:  New name for the workbook file. String.

        Returns:
            Nothing.

        """

        xfile._save_file(self._filename, self._datasourceTree, new_filename)

    @property
    def name(self):
        """ Name of the datasource. """
        return self._name

    @property
    def version(self):
        """ Version of the datasource. """
        return self._version

    @property
    def caption(self):
        """ User defined name for the datasourse. """
        return self._caption

    @caption.setter
    def caption(self, value):
        self._datasourceXML.set('caption', value)
        self._caption = value

    @caption.deleter
    def caption(self):
        del self._datasourceXML.attrib['caption']
        self._caption = ''

    @property
    def connections(self):
        """ List of connections are used in workbook. """
        return self._connections

    def clear_repository_location(self):
        tag = self._datasourceXML.find('./repository-location')
        if tag is not None:
            self._datasourceXML.remove(tag)

    @property
    def fields(self):
        """ Key-value result of field's names and its attributes. Dict. """
        if not self._fields:
            self._refresh_fields()
        return self._fields

    def _refresh_fields(self):
        self._fields = self._get_all_fields()

    def _get_all_fields(self):
        # Some columns are represented by `column` tags and others as `metadata-record` tags
        # Find them all and chain them into one dictionary
        column_field_objects = self._get_column_objects()
        existing_column_fields = [x.id for x in column_field_objects]
        metadata_only_field_objects = (x for x in self._get_metadata_objects() if x.id not in existing_column_fields)
        field_objects = itertools.chain(column_field_objects, metadata_only_field_objects)

        return FieldDictionary({k: v for k, v in field_objects})

    def _get_metadata_objects(self):
        return (_column_object_from_metadata_xml(x)
                for x in self._datasourceTree.findall(".//metadata-record[@class='column']"))

    def _get_column_objects(self):
        return [_column_object_from_column_xml(self._datasourceTree, xml)
                for xml in self._datasourceTree.findall('.//column')]

    def _get_custom_sql(self):
        return [qry for qry in self._datasourceXML.iter('relation')]

    def add_field(self, name, datatype, role, field_type, caption, hidden):
        """ Adds a base field object with the given values.

        Args:
            name: Name of the new Field. String.
            datatype:  Datatype of the new field. String.
            role:  Role of the new field. String.
            field_type:  Type of the new field. String.
            caption:  Caption of the new field. String.

        Returns:
            The new field that was created. Field.
        """
        # TODO: A better approach would be to create an empty column and then
        # use the input validation from its "Field"-object-representation to set values.
        # However, creating an empty column causes errors :(

        # If no caption is specified, create one with the same format Tableau does
        if not caption:
            caption = name.replace('[', '').replace(']', '').title()

        # Create the elements
        column = Field.create_field_xml(caption, datatype, hidden, role, field_type, name)

        self._datasourceTree.getroot().append(column)

        # Refresh fields to reflect changes and return the Field object
        self._refresh_fields()
        return self.fields[name]

    def remove_field(self, field):
        """ Remove a given field

        Args:
            field: The field to remove. ET.Element

        Returns:
            None
        """
        if not field or not isinstance(field, Field):
            raise ValueError("Need to supply a field to remove element")

        self._datasourceTree.getroot().remove(field.xml)
        self._refresh_fields()

    ###########
    # Calculations
    ###########
    @property
    def calculations(self):
        """ Returns all calculated fields.
        """
        return {k: v for k, v in self.fields.items() if v.calculation is not None}

    def add_calculation(self, caption, formula, datatype, role, type, hidden):
        """ Adds a calculated field with the given values.

        Args:
            caption:  Caption of the new calculation. String.
            formula:  Formula of the new calculation. String.
            datatype:  Datatype of the new calculation (string, integer, etc). String.
            role:  Role of the new calculation (Dimension or Measure). String.
            type:  Type of the new calculation (quantitative, ordinal, nominal). String.
            hidden:  Whether the new calculation is hidden. Boolean

        Returns:
            The new calculated field that was created. Field.
        """
        # Dynamically create the name of the field
        name = '[Calculation_{}]'.format(str(uuid4().int)[:18])
        field = self.add_field(name, datatype, role, type, caption, hidden)
        field.calculation = formula

        return field



from test.assets.index import *

import unittest
import zipfile
from tableaudocumentapi.xfile import find_file_in_zip
from tableaudocumentapi import Workbook, Datasource

TEST_ASSET_DIR = os.path.join(
    os.path.dirname(__file__),
    'assets'
)

BAD_ZIP_FILE = os.path.join(
    TEST_ASSET_DIR,
    'BadZip.zip'
)

TWBX_WITH_CACHE_FILES = os.path.join(
    TEST_ASSET_DIR,
    'Cache.twbx'
)


class XFileEdgeTests(unittest.TestCase):
    def test_find_file_in_zip_no_xml_file(self):
        badzip = zipfile.ZipFile(BAD_ZIP_FILE)
        self.assertIsNone(find_file_in_zip(badzip))

    def test_only_find_twbs(self):
        twb_from_twbx_with_cache = zipfile.ZipFile(TWBX_WITH_CACHE_FILES)
        self.assertEqual(find_file_in_zip(twb_from_twbx_with_cache), 'Superstore.twb')


class Namespacing(unittest.TestCase):

    def assertContainsUserNamespace(self, filename):
        with open(filename, 'r') as in_file:
            # the namespace is in the first five lines for all the docs I've checked
            lineCount = 0
            doc_beginning_excerpt = ""
            while lineCount < 5:
                doc_beginning_excerpt += (in_file.readline().strip())  # first line should be xml tag
                lineCount += 1
            found = doc_beginning_excerpt.rfind("xmlns:user=")
            # print(doc_beginning_excerpt[found:found+10])
            self.assertRegex(doc_beginning_excerpt, "xmlns:user=")

    def test_save_preserves_namespace_twb(self):
        filename = COMPLEX_TWB
        self.assertContainsUserNamespace(filename)
        wb = Workbook(filename)
        new_name = 'saved-as-twb.twb'
        wb.save_as(new_name)
        self.assertContainsUserNamespace(new_name)

    def demo_bug_ns_not_preserved_if_not_used(self):
        filename = TABLEAU_10_TDS
        self.assertContainsUserNamespace(filename)
        wb = Datasource.from_file(filename)
        new_name = 'saved-as-tds.tds'
        wb.save_as(new_name)
        self.assertContainsUserNamespace(new_name)





KNOWN_DB_CLASSES = ('msaccess',
                    'msolap',
                    'bigquery',
                    'asterncluster',
                    'bigsql',
                    'composite',
                    'aurora',
                    'awshadoophive',
                    'dataengine',
                    'DataStax',
                    'db2',
                    'essbase',
                    'exasolution',
                    'excel',
                    'excel-direct',
                    'excel-reader',
                    'firebird',
                    'powerpivot',
                    'genericodbc',
                    'google-analytics',
                    'googlecloudsql',
                    'google-sheets',
                    'greenplum',
                    'saphana',
                    'hadoophive',
                    'hortonworkshadoophive',
                    'maprhadoophive',
                    'marklogic',
                    'memsql',
                    'mysql',
                    'netezza',
                    'oracle',
                    'paraccel',
                    'postgres',
                    'progressopenedge',
                    'redshift',
                    'snowflake',
                    'spark',
                    'splunk',
                    'kognitio',
                    'sqlserver',
                    'salesforce',
                    'sapbw',
                    'sybasease',
                    'sybaseiq',
                    'tbio',
                    'teradata',
                    'vectorwise',
                    'vertica',
                    'denormalized-cube',
                    'csv',
                    'textscan',
                    'webdata',
                    'webdata-direct',
                    'cubeextract')


def is_valid_dbclass(dbclass):
    return dbclass in KNOWN_DB_CLASSES



import unittest
import os.path

from tableaudocumentapi import Datasource, Workbook

TEST_ASSET_DIR = os.path.join(
    os.path.dirname(__file__),
    'assets'
)
EPHEMERAL_FIELD_FILE = os.path.join(
    TEST_ASSET_DIR,
    'ephemeral_field.twb'
)

SHAPES_FILE = os.path.join(
    TEST_ASSET_DIR,
    'shapes_test.twb'
)

DASHBOARDS_FILE = os.path.join(
    TEST_ASSET_DIR,
    'filtering.twb'
)


class EphemeralFields(unittest.TestCase):
    def test_ephemeral_fields_do_not_cause_errors(self):
        wb = Workbook(EPHEMERAL_FIELD_FILE)
        self.assertIsNotNone(wb)


class Shapes(unittest.TestCase):
    def test_shape_exist(self):
        wb = Workbook(SHAPES_FILE)
        self.assertEqual(wb.shapes, ['Bug Tracking/bug.png',
                                     'Bug Tracking/icon-scheduleitem.png',
                                     'Bug Tracking/light.png',
                                     'Bug Tracking/mail.png',
                                     ]
                         )

    def test_shape_count(self):
        wb = Workbook(SHAPES_FILE)
        self.assertEqual(len(wb.shapes), 4)


class Dashboards(unittest.TestCase):
    def test_dashboards_setup(self):
        wb = Workbook(DASHBOARDS_FILE)
        self.assertIsNotNone(wb)
        self.assertEqual(wb.dashboards, ['setTest'])



import os
import os.path
import shutil
import tempfile
import unittest


from tableaudocumentapi import Datasource, Workbook

TEST_ASSET_DIR = os.path.join(
    os.path.dirname(__file__),
    'assets'
)
TEST_TDS_FILE = os.path.join(
    TEST_ASSET_DIR,
    'datasource_test.tds'
)

TEST_TWB_FILE = os.path.join(
    TEST_ASSET_DIR,
    'datasource_test.twb'
)


class DataSourceFieldsTDS(unittest.TestCase):

    def setUp(self):
        self.ds = Datasource.from_file(TEST_TDS_FILE)
        self.to_delete = set()

    def cleanUp(self):
        for path in self.to_delete:
            if os.path.isdir(path):
                shutil.rmtree(path, ignore_errors=True)
            elif os.path.isfile(path):
                os.unlink(path)

    def get_temp_file(self, filename):
        tempdir = tempfile.mkdtemp('tda-datasource')
        self.to_delete.add(tempdir)
        return os.path.join(tempdir, filename)

    def test_datasource_returns_correct_fields(self):
        self.assertIsNotNone(self.ds.fields)
        self.assertIsNotNone(self.ds.fields.get('[Number of Records]', None))

    def test_datasource_returns_calculation_from_fields(self):
        self.assertEqual('1', self.ds.fields['[Number of Records]'].calculation)

    def test_datasource_uses_metadata_record(self):
        self.assertEqual('Sum', self.ds.fields['[x]'].default_aggregation)

    def test_datasource_column_name_contains_apostrophy(self):
        self.assertIsNotNone(self.ds.fields.get("[Today's Date]", None))

    def test_datasource_field_can_get_caption(self):
        self.assertEqual(self.ds.fields['[a]'].caption, 'A')
        self.assertEqual(getattr(self.ds.fields['[a]'], 'caption', None), 'A')

    def test_datasource_field_caption_can_be_used_to_query(self):
        self.assertIsNotNone(self.ds.fields.get('A', None))

    def test_datasource_field_is_nominal(self):
        self.assertTrue(self.ds.fields['[a]'].is_nominal)

    def test_datasource_field_is_quantitative(self):
        self.assertTrue(self.ds.fields['[y]'].is_quantitative)

    def test_datasource_field_is_ordinal(self):
        self.assertTrue(self.ds.fields['[x]'].is_ordinal)

    def test_datasource_field_datatype(self):
        self.assertEqual(self.ds.fields['[x]'].datatype, 'integer')

    def test_datasource_field_role(self):
        self.assertEqual(self.ds.fields['[x]'].role, 'measure')

    def test_datasource_field_description(self):
        actual = self.ds.fields['[a]'].description
        self.assertIsNotNone(actual)
        self.assertTrue(u'muted gray' in actual)

    def test_datasource_caption(self):
        actual = self.ds.caption
        self.assertIsNotNone(actual)
        self.assertEqual(actual, 'foo')

    def test_datasource_can_set_caption(self):
        filename = self.get_temp_file('test_datasource_can_set_caption')
        self.ds.caption = 'bar'
        self.ds.save_as(filename)

        actual = Datasource.from_file(filename)
        self.assertIsNotNone(actual)
        self.assertIsNotNone(actual.caption)
        self.assertEqual(actual.caption, 'bar')

    def test_datasource_can_remove_caption(self):
        filename = self.get_temp_file('test_datasource_can_remove_caption')
        del self.ds.caption
        self.ds.save_as(filename)

        actual = Datasource.from_file(filename)
        self.assertIsNotNone(actual)
        self.assertEqual(actual.caption, '')

    def test_datasource_clear_repository_location(self):
        filename = os.path.join(TEST_ASSET_DIR, 'clear-repository-test.tds')

        self.assertIsNotNone(self.ds._datasourceXML.find('.//repository-location'))
        self.ds.clear_repository_location()
        try:
            self.ds.save_as(filename)
            with open(filename, 'r') as newfile:
                self.assertFalse('repository-location' in newfile.read())
        finally:
            if os.path.exists(filename):
                os.unlink(filename)


class DataSourceFieldsTWB(unittest.TestCase):

    def setUp(self):
        self.wb = Workbook(TEST_TWB_FILE)
        # Assume the first datasource in the file
        self.ds = self.wb.datasources[0]

    def test_datasource_fields_loaded_in_workbook(self):
        self.assertIsNotNone(self.ds.fields)
        self.assertIsNotNone(self.ds.fields.get('[Number of Records]', None))


class DataSourceFieldsFoundIn(unittest.TestCase):

    def setUp(self):
        self.wb = Workbook(TEST_TWB_FILE)
        # Assume the first datasource in the file
        self.ds = self.wb.datasources[0]

    def test_datasource_fields_found_in_returns_fields(self):
        actual_values = self.ds.fields.used_by_sheet('Sheet 1')
        self.assertIsNotNone(actual_values)
        self.assertEqual(1, len(actual_values))
        self.assertIn('A', (x.name for x in actual_values))

    def test_datasource_fields_found_in_does_not_return_fields_not_used_in_worksheet(self):
        actual_values = self.ds.fields.used_by_sheet('Sheet 1')
        self.assertIsNotNone(actual_values)
        self.assertEqual(1, len(actual_values))
        self.assertNotIn('X', (x.name for x in actual_values))

    def test_datasource_fields_found_in_returns_multiple_fields(self):
        actual_values = self.ds.fields.used_by_sheet('Sheet 2')
        self.assertIsNotNone(actual_values)
        self.assertEqual(2, len(actual_values))
        self.assertIn('A', (x.name for x in actual_values))
        self.assertIn('X', (x.name for x in actual_values))
        self.assertNotIn('Y', (x.name for x in actual_values))

    def test_datasource_fields_found_in_accepts_lists(self):
        actual_values = self.ds.fields.used_by_sheet(['Sheet 1', 'Sheet 2'])
        self.assertIsNotNone(actual_values)
        self.assertEqual(2, len(actual_values))
        self.assertIn('A', (x.name for x in actual_values))
        self.assertIn('X', (x.name for x in actual_values))
        self.assertNotIn('Y', (x.name for x in actual_values))



import functools
from lxml import etree as ET
from xml.dom import minidom

from tableaudocumentapi.property_decorators import argument_is_one_of

_ATTRIBUTES = [
    'id',           # Name of the field as specified in the file, usually surrounded by [ ]
    'caption',      # Name of the field as displayed in Tableau unless an aliases is defined
    'datatype',     # Type of the field within Tableau (string, integer, etc)
    'role',         # Dimension or Measure
    'type',         # three possible values: quantitative, ordinal, or nominal
    'alias',        # Name of the field as displayed in Tableau if the default name isn't wanted
    'calculation',  # If this field is a calculated field, this will be the formula
    'description',  # If this field has a description, this will be the description (including formatting tags)
    'hidden',       # If this field has been hidden
]

_METADATA_ATTRIBUTES = [
    'aggregation',  # The type of aggregation on the field (e.g Sum, Avg)
]

_METADATA_TO_FIELD_MAP = [
    ('local-name', 'id'),
    ('local-type', 'datatype'),
    ('remote-alias', 'alias')
]


def _find_metadata_record(record, attrib):
    element = record.find('.//{}'.format(attrib))
    if element is None:
        return None
    return element.text


class Field(object):
    """ Represents a field in a datasource """

    def __init__(self, column_xml=None, metadata_xml=None):

        # Initialize all the possible attributes
        for attrib in _ATTRIBUTES:
            setattr(self, '_{}'.format(attrib), None)
        for attrib in _METADATA_ATTRIBUTES:
            setattr(self, '_{}'.format(attrib), None)
        self._worksheets = set()

        if column_xml is not None:
            self._initialize_from_column_xml(column_xml)
            self._xml = column_xml
            # This isn't currently called because of the way we get the data from the xml,
            # but during the refactor, we might need it.  This is commented out as a reminder
            # if metadata_xml is not None:
            #     self.apply_metadata(metadata_xml)

        elif metadata_xml is not None:
            self._xml = metadata_xml
            self._initialize_from_metadata_xml(metadata_xml)

        else:
            raise AttributeError('column_xml or metadata_xml needed to initialize field')

    def _initialize_from_column_xml(self, xmldata):
        for attrib in _ATTRIBUTES:
            self._apply_attribute(xmldata, attrib, lambda x: xmldata.attrib.get(x, None))

    def _initialize_from_metadata_xml(self, xmldata):
        for metadata_name, field_name in _METADATA_TO_FIELD_MAP:
            self._apply_attribute(xmldata, field_name,
                                  lambda x: getattr(xmldata.find('.//{}'.format(metadata_name)), 'text', None),
                                  read_name=metadata_name)
            print(metadata_name, field_name)
        self.apply_metadata(xmldata)

    @classmethod
    def create_field_xml(cls, caption, datatype, hidden, role, field_type, name):
        column = ET.Element('column')
        column.set('caption', caption)
        column.set('datatype', datatype)
        column.set('hidden', hidden)
        column.set('role', role)
        column.set('type', field_type)
        column.set('name', name)
        return column

    ########################################
    # Special Case methods for construction fields from various sources
    # not intended for client use
    ########################################
    def apply_metadata(self, metadata_record):
        for attrib in _METADATA_ATTRIBUTES:
            self._apply_attribute(metadata_record, attrib, functools.partial(_find_metadata_record, metadata_record))

    def add_used_in(self, name):
        self._worksheets.add(name)

    @classmethod
    def from_column_xml(cls, xmldata):
        return cls(column_xml=xmldata)

    @classmethod
    def from_metadata_xml(cls, xmldata):
        return cls(metadata_xml=xmldata)

    def _apply_attribute(self, xmldata, attrib, default_func, read_name=None):
        if read_name is None:
            read_name = attrib
        if hasattr(self, '_read_{}'.format(read_name)):
            value = getattr(self, '_read_{}'.format(read_name))(xmldata)
        else:
            value = default_func(attrib)

        setattr(self, '_{}'.format(attrib), value)

    @property
    def name(self):
        """ Provides a nice name for the field which is derived from the alias, caption, or the id.

        The name resolves as either the alias if it's defined, or the caption if alias is not defined,
        and finally the id which is the underlying name if neither of the fields exist. """
        alias = getattr(self, 'alias', None)
        if alias:
            return alias

        caption = getattr(self, 'caption', None)
        if caption:
            return caption

        return self.id

    @property
    def id(self):
        """ Name of the field as specified in the file, usually surrounded by [ ] """
        return self._id

    @property
    def xml(self):
        """ XML representation of the field. """
        return self._xml

    def pretty_xml(self):
        """Return a pretty-printed XML string for the Element.
        """
        rough_string = ET.tostring(self._xml, 'utf-8')
        prepared_string = minidom.parseString(rough_string)
        print_string = prepared_string.toprettyxml(indent="  ", newl="")
        return print_string.lstrip('<?xml version="1.0" ?>')

    def __str__(self):
        """ String representation of the field (only includes usable attributes) """
        # TODO: ideally this should just loop through the ATTRIBUTES so it doesn't need touching for new ones
        output = "------ FIELD {} ({}/{}/{}): {}(type), {}(datatype), {}(role), {}(aggregation)".format(
            self.name, self.caption, self.alias, self.id, self.type, self.datatype, self.role, self.default_aggregation)
        return output

    def detailed_str(self):
        if self.calculation:
            calc = "\ncalc: `{}`".format(self.calculation)
        else:
            calc = ""

    ########################################
    # Attribute getters and setters
    ########################################

    @property
    def caption(self):
        """ Name of the field as displayed in Tableau unless an aliases is defined """
        return self._caption

    @caption.setter
    def caption(self, caption):
        """ Set the caption of a field

            Args:
                caption:  New caption. String.

            Returns:
                Nothing.
        """
        self._caption = caption
        self._xml.set('caption', caption)

    @property
    def alias(self):
        """ Name of the field as displayed in Tableau if the default name isn't wanted """
        return self._alias

    @alias.setter
    def alias(self, alias):
        """ Set the alias of a field

            Args:
                alias:  New alias. String.

            Returns:
                Nothing.
        """
        self._alias = alias
        self._xml.set('alias', alias)

    @property
    def datatype(self):
        """ Type of the field within Tableau (string, integer, etc) """
        return self._datatype

    @datatype.setter
    @argument_is_one_of('string', 'integer', 'date', 'boolean')
    def datatype(self, datatype):
        """ Set the datatype of a field

            Args:
                datatype:  New datatype. String.

            Returns:
                Nothing.
        """
        self._datatype = datatype
        self._xml.set('datatype', datatype)

    @property
    def hidden(self):
        """ If the column is Hidden ('true', 'false') """
        return self._hidden

    @hidden.setter
    @argument_is_one_of('true', 'false')
    def hidden(self, hidden):
        """ Set the hidden property of a field

            Args:
                hidden:  New hidden. String.

            Returns:
                Nothing.
        """
        self._hidden = hidden
        self._xml.set('hidden', hidden)

    @property
    def role(self):
        """ Dimension or Measure """
        return self._role

    @role.setter
    @argument_is_one_of('dimension', 'measure')
    def role(self, role):
        """ Set the role of a field

            Args:
                role:  New role. String.

            Returns:
                Nothing.
        """
        self._role = role
        self._xml.set('role', role)

    @property
    def type(self):
        """ Type of field (quantitative, ordinal, nominal) """
        return self._type

    @type.setter
    @argument_is_one_of('quantitative', 'ordinal', 'nominal')
    def type(self, field_type):
        """ Set the type of a field

            Args:
                field_type:  New type. String.

            Returns:
                Nothing.
        """
        self._type = field_type
        self._xml.set('type', field_type)

    ########################################
    # Aliases getter and setter
    # Those are NOT the 'alias' field of the column,
    # but instead the key-value aliases in its child elements
    ########################################

    def add_alias(self, key, value):
        """ Add an alias for a given display value.

            Args:
                key:  The data value to map. Example: "1". String.
                value: The display value for the key. Example: "True". String.
            Returns:
                Nothing.
        """

        # determine whether there already is an aliases-tag
        aliases = self._xml.find('aliases')
        # and create it if there isn't
        if not aliases:  # ignore the FutureWarning, does not apply to our usage
            aliases = ET.Element('aliases')
            self._xml.append(aliases)

        # find out if an alias with this key already exists and use it
        existing_alias = [tag for tag in aliases.findall('alias') if tag.get('key') == key]
        # if not, create a new ET.Element
        alias = existing_alias[0] if existing_alias else ET.Element('alias')

        alias.set('key', key)
        alias.set('value', value)
        if not existing_alias:
            aliases.append(alias)

    @property
    def aliases(self):
        """ Returns all aliases that are registered under this field.

        Returns:
            Key-value mappings of all registered aliases. Dict.
        """
        aliases_tag = self._xml.find('aliases') or []    # ignore the FutureWarning, does not apply to our usage
        return {a.get('key', 'None'): a.get('value', 'None') for a in list(aliases_tag)}

    ########################################
    # Attribute getters
    ########################################

    @property
    def is_quantitative(self):
        """ A dependent value, usually a measure of something

        e.g. Profit, Gross Sales """
        return self.type == 'quantitative'

    @property
    def is_ordinal(self):
        """ Is this field a categorical field that has a specific order

        e.g. How do you feel? 1 - awful, 2 - ok, 3 - fantastic """
        return self.type == 'ordinal'

    @property
    def is_nominal(self):
        """ Is this field a categorical field that does not have a specific order

        e.g. What color is your hair? """
        return self.type == 'nominal'

    @property
    def calculation(self):
        """ If this field is a calculated field, this will be the formula """
        return self._calculation

    @calculation.setter
    def calculation(self, new_calculation):
        """ Set the calculation of a calculated field.

        Args:
            new_calculation: The new calculation/formula of the field. String.
        """
        if self.calculation is None:
            calculation = ET.Element('calculation')
            calculation.set('class', 'tableau')
            calculation.set('formula', new_calculation)
            # Append the elements to the respective structure
            self._xml.append(calculation)

        else:
            self._xml.find('calculation').set('formula', new_calculation)

        self._calculation = new_calculation

    @property
    def default_aggregation(self):
        """ The default type of aggregation on the field (e.g Sum, Avg)"""
        return self._aggregation

    @property
    def description(self):
        """ The contents of the <desc> tag on a field """
        return self._description

    @property
    def worksheets(self):
        """ Worksheets which uses field. """
        return list(self._worksheets)

    ######################################
    # Special Case handling methods for reading the values from the XML
    ######################################
    @staticmethod
    def _read_id(xmldata):
        # ID is actually the name of the field, but to provide a nice name, we call this ID
        return xmldata.attrib.get('name', None)

    @staticmethod
    def _read_calculation(xmldata):
        # The formula for a calculation is stored in a child element, so we need to pull it out separately.
        calc = xmldata.find('.//calculation')
        if calc is None:
            return None

        return calc.attrib.get('formula', None)

    @staticmethod
    def _read_description(xmldata):
        description = xmldata.find('.//desc')
        if description is None:
            return None

        description_string = ET.tostring(description, encoding='utf-8')
        # Format expects a unicode string so in Python 2 we have to do the explicit conversion
        if isinstance(description_string, bytes):
            description_string = description_string.decode('utf-8')

        return description_string



from setuptools import setup
    
setup(
    name='tableaudocumentapi',
    version='0.10',
    author='Tableau',
    author_email='github@tableau.com',
    url='https://github.com/tableau/document-api-python',
    packages=['tableaudocumentapi'],
    license='MIT',
    description='A Python module for working with Tableau files.',
    long_description="file: README.md",
    long_description_content_type="text/markdown",
    test_suite='test',
    install_requires=['lxml']
)



from functools import wraps


def argument_is_one_of(*allowed_values):
    def property_type_decorator(func):
        @wraps(func)
        def wrapper(self, value):
            if value not in allowed_values:
                error = "Invalid argument: {0}. {1} must be one of {2}."
                msg = error.format(value, func.__name__, allowed_values)
                raise ValueError(error)
            return func(self, value)

        return wrapper

    return property_type_decorator



import weakref


_no_default_value = object()


def _resolve_value(key, value):
    retval = None
    try:
        if hasattr(value, 'get'):
            retval = value.get(key, None)

        if retval is None:
            retval = getattr(value, key, None)
    except AttributeError:
        # We should never hit this.
        retval = None
    return retval


def _build_index(key, d):
    return {_resolve_value(key, v): k
            for k, v in d.items()
            if _resolve_value(key, v) is not None}


# TODO: Improve this to be more generic
class MultiLookupDict(dict):
    def __init__(self, args=None):
        if args is None:
            args = {}
        super(MultiLookupDict, self).__init__(args)
        self._indexes = {
            'alias': weakref.WeakValueDictionary(),
            'caption': weakref.WeakValueDictionary()
        }
        self._populate_indexes()

    def _populate_indexes(self):
        self._indexes['alias'] = _build_index('alias', self)
        self._indexes['caption'] = _build_index('caption', self)

    def _get_real_key(self, key):
        if key in self._indexes['alias']:
            return self._indexes['alias'][key]
        if key in self._indexes['caption']:
            return self._indexes['caption'][key]

        return key

    def __setitem__(self, key, value):
        real_key = self._get_real_key(key)

        dict.__setitem__(self, real_key, value)

    def get(self, key, default_value=_no_default_value):
        try:
            return self[key]
        except KeyError:
            if default_value is not _no_default_value:
                return default_value
            raise

    def __getitem__(self, key):
        real_key = self._get_real_key(key)
        return dict.__getitem__(self, real_key)
